/*
** unrarhp.c - recover passwd of rar files encrypted with "rar a -hp[passwd]"
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "sha1.hpp"
#include "rijndael.hpp"

#define MAXPASSWD	128

char	*ucs2_str(char *dst, const char *src)
{
    char	*ret;

    ret = dst;
    while (*src)
      {
	*dst++ = *src++;
	*dst++ = '\0';
      }
    return ret;
}

int	ucs2_len(const char *s)
{
    return 2 * strlen(s);
}

#define PAT_ANYCHAR	'?'
#define PAT_CHARS	"0123456789""abcdefghijklmnopqrstuvwxyz"

/*
** Increment the char pointed to by 'c', according to PAT_CHARS.
**
** Return 1 if the increment has generated a carry, else return 0.
*/
int	inc_chr(char *c)
{
    if (*c >= '0' && *c < '9')
	(*c)++;
    else if (*c == '9')
	(*c) = 'a';
    else if (*c >= 'a' && *c < 'z')
	(*c)++;
    else if (*c == 'z')
      {
	(*c) = '0';
	return 1;
      }
    else
	fprintf(stderr, "Bug: c == '%c'\n", *c), exit(1);
    return 0;
}

/*
** Generate the next password, based on a pattern.
**
**   @init     if set, initialize the pwd to the 1st possible value
**   @pwd      passwd is written in this buffer
**   @pat      pattern
**
** Return 1 if a passwd has been generated or 0 if we have reached the
** end of possible passwords.
*/
int	pwd_next(int init, char *pwd, const char *pat)
{
    int		i;

    if (init)
      {
	strcpy(pwd, pat);
	for (i = 0; pwd[i]; i++)
	    if (pwd[i] == PAT_ANYCHAR)
		pwd[i] = PAT_CHARS[0];
      }
    else
      {
	for (i = strlen(pwd) - 1; i >= 0; i--)
	    if (pat[i] == PAT_ANYCHAR)
		if (!inc_chr(&pwd[i]))
		    break ;
	if (i == -1)
	    return 0;
      }
    return 1;
}

/*
** Borrowed from sources of unrar
*/
void	gen_aes_val(unsigned char *aes_key, unsigned char *aes_iv,
	char *pwd, int plen, unsigned char *salt)
{
    unsigned char	RawPsw[plen + 8];
    int			RawLength = plen + 8;
    hash_context	c;

    memcpy(RawPsw, pwd, plen);
    memcpy(RawPsw + plen, salt, 8);

    hash_initial(&c);
    const int HashRounds=0x40000;
    for (int I=0;I<HashRounds;I++)
    {
      hash_process( &c, RawPsw, RawLength);
      byte PswNum[3];
      PswNum[0]=(byte)I;
      PswNum[1]=(byte)(I>>8);
      PswNum[2]=(byte)(I>>16);
      hash_process( &c, PswNum, 3);
      if (I%(HashRounds/16)==0)
      {
        hash_context tempc=c;
        uint32 digest[5];
        hash_final( &tempc, digest);
        aes_iv[I/(HashRounds/16)]=(byte)digest[4];
      }
    }
    uint32 digest[5];
    hash_final( &c, digest);
    for (int I=0;I<4;I++)
      for (int J=0;J<4;J++)
        aes_key[I*4+J]=(byte)(digest[I]>>(J*8));
}

/*
** Test a password, using the end-of-archive block.
**
**   @salt point to 8 bytes of salt data
**   @in is a 16-byte encrypted block representing the end-of-archive block
**   @p is a password to try
**
** Return true on good passwd, false on bad passwd.
*/
int	test_password(unsigned char *salt, unsigned char *in, const char *p)
{
    unsigned char	aes_key[16];
    unsigned char	aes_iv[16];
    Rijndael		rin;
    unsigned char	out[16];
    char		p2[MAXPASSWD];

    gen_aes_val(aes_key, aes_iv, ucs2_str(p2, p), ucs2_len(p), salt);
    rin.init(Rijndael::Decrypt, aes_key, aes_iv);
    rin.blockDecrypt(in, 16, out);
    /* cmp with the usual end-of-archive block present in (all?) rar files */
    return !memcmp(out, "\xc4\x3d\x7b\x00\x40\x07\x00", 7);
}

void	disp_stats(int i_cur, int end)
{
    static int			i_old = -1;
    static struct timeval	tv0;
    static struct timeval	tv1;
    double			f;

    if (!i_cur)
      {
	if (-1 == gettimeofday(&tv0, NULL))
	    perror("gettimeofday"), exit(1);
	f = 0;
      }
    else
      {
	if (-1 == gettimeofday(&tv1, NULL))
	    perror("gettimeofday"), exit(1);
	f =
	    ((double)tv1.tv_sec + (double)tv1.tv_usec / 1.e6) -
	    ((double)tv0.tv_sec + (double)tv0.tv_usec / 1.e6);
	tv0 = tv1;
	f = (i_cur - i_old) / f;
      }
    fprintf(stderr, "\r%d pwds, %.2f pwd/sec ", i_cur, f);
    if (end)
	fprintf(stderr, "\n");
    i_old = i_cur;
}

void	found_it(const char *pwd)
{
    int	i;

    printf("| passwd is: %s (hex dump:", pwd);
    for (i = 0; i < (int)strlen(pwd); i++)
	printf(" %02x", (unsigned char)pwd[i]);
    printf(")\n");
    exit(0);
}

void	loop_passwords(unsigned char *salt, unsigned char *in, const char *pat)
{
    char	*p;
    int		i;
    int		first;

    if (!(p = (char *)malloc(strlen(pat) + 1)))
	perror("malloc"), exit(1);
    p[0] = 0;
    i = 0;
    first = 1;
    while (pwd_next(first, p, pat))
      {
	first = 0;
	if (!(i % 41))
	    disp_stats(i, 0);
	/*fprintf(stderr, "\ntrying pwd: %s\n", p);*/
	i++;
	if (test_password(salt, in, p))
	  {
	    disp_stats(i, 0);
	    found_it(p);
	  }
      }
    disp_stats(i, 1);
    free(p);
}

void	do_file(const char *f, const char *pat)
{
    unsigned char	buf[24];
    int		fd;
    int		n;

    if (-1 == (fd = open(f, O_RDONLY, 0)))
	perror(f), exit(1);
    if (-1 == lseek(fd, -24, SEEK_END))
	perror("lseek"), exit(1);
    n = read(fd, buf, sizeof (buf));
    if (!n)
	fprintf(stderr, "Premature end-of-file\n"), exit(1);
    if (n != sizeof (buf))
	fprintf(stderr, "Read only %d bytes instead of %d\n",
		n, (int)sizeof (buf)), exit(1);
    close(fd);
    loop_passwords(buf, buf + 8, pat);
}

void	usage(void)
{
    fprintf(stderr,
	    "Usage: unrarhp {file} {pattern}\n"
	    "  {file}     rar file to crack\n"
	    "  {pattern}  passwd pattern to test ('%c' for unknown char)\n"
	    , PAT_ANYCHAR
	   );
}

int	main(int argc, char **argv)
{
    const char	*file;
    const char	*pattern;

    if (argc == 3)
      {
	file = argv[1];
	pattern = argv[2];
      }
    else
	usage(), exit(1);
    do_file(file, pattern);
    return 1; /* if we reach this point, passwd has not been found */
}
